//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================

#include "actors.hh"
#include "player.hh"
#include "sound.hh"
#include "world.hh"
#include "enigma.hh"

using namespace actors;
using px::V3;

//======================================================================
// ACTORS
//======================================================================

Actor::Actor(const string &kind, const px::V3 &p)
    : Object(kind), actorinfo(p, V3())
{
    set_attrib("mouseforce", 0.0);
    sprite_id = 0;
}

void 
Actor::set_attrib(const string &key, const Value &val)
{
    if (key == "player")
    {
        int iplayer = to_int(val);
        player::SetMainActor(iplayer, this);
    }
    Object::set_attrib(key, val);
}

void 
Actor::on_creation(const px::V3 &p)
{
    set_model(get_kind(), p);
}

enigma::GridPos get_field(px::V3 p)
{
    return enigma::GridPos(static_cast<int>(p[0]), static_cast<int>(p[1]));
}

void 
Actor::move()
{
    using namespace world;
    GridPos  field = get_field(actorinfo.pos);
    GridPos ofield = get_field(actorinfo.oldpos);

    if (field != ofield) {
        if (Floor *fl = GetFloor(field))
            fl->actor_enter(this);
        if (Floor *ofl = GetFloor(ofield))
            ofl->actor_leave(this);

        if (Item *it = GetItem(field))
            it->actor_enter(this);
        if (Item *oit = GetItem(ofield))
            oit->actor_leave(this);
    }
    if (Floor *fl = GetFloor(field))
        fl->actor_hit(this);

    if (Item *it = GetItem(field))
        if (it->actor_hit(this))
            player::PickupItem(this, field);

    if (Stone *st = GetStone(field))
        st->actor_inside(this);

    display::MoveSprite(sprite_id, actorinfo.pos);
    on_motion(actorinfo.pos);
}

void 
Actor::set_model(const string & mname, const px::V3 &pos)
{
    sprite_id = AddSprite(display::SPRITE_ACTOR, pos, mname);
}

void 
Actor::set_model(const string &name)
{
    display::ReplaceSprite(sprite_id, name);
}



//----------------------------------------
// Rotor
//----------------------------------------
namespace
{
    class Rotor : public Actor {
    public:
        Rotor();
    private:
        // Actor interface.
        bool is_dead() { return false; }

        // Object interface.
        Object* clone() { return new Rotor; }
        void dispose() { delete this; }
    };
}

Rotor::Rotor()
    : Actor("ac-rotor", V3())
{
    world::ActorInfo *ai = get_actorinfo();
    ai->radius = 20/64.0;
    ai->mass = 0.8;
}

//----------------------------------------
// BasicBall
//----------------------------------------
namespace
{
    class BasicBall : public Actor
    {
    protected:
        BasicBall(const char *kind, double radius, double mass)
            : Actor(kind, V3())
            , state(NORMAL)
        {
            world::ActorInfo *ai = get_actorinfo();
            ai->radius = radius;
            ai->mass = mass;
        }

        enum State { NORMAL, SHATTERING, DROWNING, FALLING, JUMPING, DEAD };
        State state;

        void set_model_cb(const string &m) { 
            set_model(m);
            display::SetCallback(get_spriteid(),
                                 make_cb(this, &BasicBall::animcb));
        }
        virtual void change_state(State newstate) = 0;
        virtual void animcb() {}

        // Actor interface.
        bool is_dead() { return state == DEAD; }
        bool is_flying() { return state == JUMPING; }

        // Object interface.
        void message(const string &m, const Value &);
    };
}

void 
BasicBall::message(const string &m, const Value &)
{
    if (state == NORMAL) {
        if (m == "shatter")
            change_state(SHATTERING);
        else if (m == "fall") 
            change_state(FALLING);
        else if (m == "drown")
            change_state(DROWNING);
        else if (m == "jump")
            change_state(JUMPING);
    }
}


//----------------------------------------
// BlackBall
//----------------------------------------

/** \page ac-blackball Black Ball Actor

This is the black marble you control most of the time.

\subsection blackballm Messages

- \b shatter Destroy the marble by shattering it.
- \b fall Destroy the marble by pulling it into the abyss.

*/
namespace
{
    class BlackBall : public BasicBall {
    public:
        BlackBall();
    private:
        void on_creation(const V3 &p)
        {
            set_model("ac-blackball", p);
            display::FollowSprite(get_spriteid());
        }

        void change_state(State newstate);
        void animcb();

        Object* clone() { return new BlackBall; }
        void dispose() { delete this; }
    };
}

BlackBall::BlackBall() 
    : BasicBall("ac-blackball", 21.0/64.0, /*mass*/ 1.0)
{
    set_attrib("mouseforce", Value(true));
    set_attrib("color", Value(0.0));
}

void 
BlackBall::change_state(State newstate)
{
    if (newstate == state)
        return;

    switch (newstate) {
    case NORMAL: break;
    case SHATTERING:
        sound::PlaySound("shatter");
        world::GrabActor(this);
        set_model_cb("ac-blackball-shatter");
        break;
    case DROWNING:
        world::GrabActor(this);
        sound::PlaySound("drown");
        set_model_cb("ring-anim");
        break;
    case FALLING:
        set_attrib("mouseforce", Value(0.0));
        world::GrabActor(this);
        set_model_cb("ac-blackball-fall");
        break;
    case DEAD: break;
    case JUMPING: 
        set_model_cb("ac-blackball-jump");
        break;
    }
    state = newstate;
}

void 
BlackBall::animcb() 
{
    switch (state) {
    case SHATTERING:
        set_model("ac-blackball-shattered");
        change_state(DEAD);
        break;
    case FALLING:
        sound::PlaySound("shatter");
        set_model("ac-blackball-fallen");
        change_state(DEAD);
        break;
    case DROWNING:
        set_model("invisible");
        change_state(DEAD);
        break;
    case JUMPING:
        set_model("ac-blackball");
        change_state(NORMAL);
        break;
    default:
        break;
    }
}

//----------------------------------------
// WhiteBall_Small
//----------------------------------------
namespace
{
    class WhiteBall_Small : public BasicBall {
    public:
        WhiteBall_Small()
            : BasicBall("ac-whiteball-small", 13/64.0, 0.7)
        {}
    private:
        void change_state(State newstate);
        void animcb();


        Object *clone() { return new WhiteBall_Small; }
        void dispose() { delete this; }
    };
}

void 
WhiteBall_Small::change_state(State newstate)
{
    switch (newstate) {
    case NORMAL: break;
    case SHATTERING:
        sound::PlaySound("shatter");
        world::GrabActor(this);
        set_model_cb("ac-whiteball-small-shatter");
        break;
    case DROWNING:
        world::GrabActor(this);
        sound::PlaySound("drown");
        set_model_cb("ring-anim");
        break;
    case FALLING:
        set_attrib("mouseforce", Value(0.0));
        world::GrabActor(this);
        set_model_cb("ac-whiteball-small-fall");
        break;
    case DEAD: break;
    case JUMPING: break;
    }
    state = newstate;
}

void 
WhiteBall_Small::animcb() 
{
    switch (state) {
    case SHATTERING:
        set_model("ac-whiteball-small-shattered");
        change_state(DEAD);
        break;
    case FALLING:
        sound::PlaySound("shatter");
        set_model("ac-blackball-fallen");
        change_state(DEAD);
        break;
    case DROWNING:
        set_model("invisible");
        change_state(DEAD);
        break;
    default:
        break;
    }
}

void 
actors::Init()
{
    using world::Register;

    Register(new Rotor);
    Register(new BlackBall);
    Register(new WhiteBall_Small);
}
