//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#ifndef DISPLAY_HH
#define DISPLAY_HH

#include "fwd.hh"
#include "enigma.hh"
#include "player.hh"
#include "px/tools.hh"
#include "px/math.hh"
#include "px/video.hh"
#include "px/callback.hh"
#include <string>

namespace display
{
    using std::string;

    class ModelCallback {
    public:
        virtual void animation_done() = 0;
    };

    // Models can be prompted to invoke a callback on completion.
    // Please note that you may not delete or replace models other
    // than the one that induced the callback from inside the callback
    // -- use a timer or a flag to do this.
    class Model {
    public:
        virtual ~Model() {}
        virtual double duration() const { return 0; }
        virtual void set_callback(px::Callback *cb) {}
        virtual void reverse() {}
    };

    class Animation {
    public:
        virtual double duration() const = 0;
        virtual void set_callback(px::Callback *cb) = 0;
        virtual void reverse() = 0;
    };

    Model * MakeModel(const string &name);

}

//
// ``Grid'' models
//
namespace display
{
    using enigma::GridPos;
    using enigma::GridLayer;
    using enigma::GridLoc;

    Model*  SetModel (const GridLoc & l, const string &modelname);
    Model*  SetModel (const GridLoc & l, Model *m);
    void    KillModel (const GridLoc & l);
    Model*  GetModel (const GridLoc & l);
    Model*  YieldModel (const GridLoc & l);
}

//
// Sprites (movable models)
//
namespace display
{
    enum SpriteLayer
    {
        SPRITE_ACTOR, SPRITE_EFFECT, SPRITE_COUNT
    };

    typedef unsigned int SpriteId;

    SpriteId AddSprite (SpriteLayer l, const px::V3 &pos, const string &modelname);
    void MoveSprite (SpriteId id, const px::V3 &newpos);
    void ReplaceSprite(SpriteId id, const string &modelname);
    void KillSprite (SpriteId id);
    Model* GetModel(SpriteId id);

    void SetCallback(SpriteId id, px::Callback *cb);

    //
    // Sprite following
    // 
    enum FollowMode { 
        FOLLOW_NONE,
        FOLLOW_SCROLLING,       // scroll the screen
        FOLLOW_SCREEN,          // flip the screen region
        FOLLOW_HALFSCREEN
    };
    void FollowSprite(SpriteId id);
    void SetFollowMode(FollowMode m);
}

namespace display
{
    enum ShadowType 
    {
        SHADOW_NONE,
        SHADOW_STIPPLE,
        SHADOW_ALPHA
    };
    void SetShadowType(ShadowType st);


    enum DisplayFlags 
    {
        SHOW_FLOOR     = 0x01,
        SHOW_STONES    = 0x02,
        SHOW_ITEMS     = 0x04,
        SHOW_SHADES    = 0x08,
        SHOW_SPRITES   = 0x10,
	SHOW_TIME      = 0x20,
	SHOW_INVENTORY = 0x40,
        SHOW_ALL       = 0x7f
    };

    void ToggleFlag(DisplayFlags flag);


    void Init();
    void Shutdown();

    // Reload all models.  This is similar to calling Init() and
    // Shutdown(), but it restarts the current level.  [This is more
    // or less a hack to make incremental changes to the models
    // easier.]
    void ReloadModels();

    void NewWorld (int w, int h);
    void RedrawAll(px::Screen *screen);
    void Redraw(px::Screen *screen, bool updatescreenp = true);
    void Tick(double dtime);

    //
    // Inventory management
    //
    using player::Inventory;
    void SetInventory(Inventory *inv);
    void RedrawInventory(Inventory *inv);
    
    //
    // Text display
    // 
    enum TextMode {
        TEXT_SCROLLING,         // scroll text from right to left
        TEXT_2SECONDS,          // show text centered, for 2 seconds
        TEXT_STATIC             // show text centered, indefinitely
    };
    void ShowText(const std::string &str, TextMode m=TEXT_SCROLLING);
}
#endif
