//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================

#include "gui.hh"
#include "enigma.hh"
#include "sound.hh"
#include "video.hh"
#include "px/px.hh"
#include <memory>

using namespace gui;
using namespace px;

//----------------------------------------
// Widget implementation.
//----------------------------------------
void 
Widget::redraw() 
{ 
    if (menu) menu->invalidate_area(get_area());
}

//----------------------------------------
// Button implementation
//----------------------------------------
void 
Button::draw(px::GC &gc) 
{
    char bordername[50];
    sprintf(bordername, "b%dx%d%s", get_w(), get_h(),
            highlighted ? "h" : "");
    px::Surface &s = enigma::GetImage(bordername);
    blit(gc, get_x(), get_y(), &s);
}



//----------------------------------------
// TextButton implementation.
//----------------------------------------
px::Font *TextButton::menufont = 0;
px::Font *TextButton::menufont_pressed = 0;

TextButton::TextButton(const string &t, ActionListener *al) 
    : text(t), pressed(false), listener(al)
{
    if (menufont == 0)
    {
        menufont = enigma::LoadFont("menufont");
        menufont_pressed = enigma::LoadFont("menufontsel");
    }
}

void 
TextButton::set_text(const std::string &t)
{
    text = t;
    redraw();
}

void 
TextButton::on_event(const SDL_Event &e)
{
    switch (e.type) {
    case SDL_MOUSEBUTTONDOWN:
        pressed=true; 
        redraw();
        break;
    case SDL_MOUSEBUTTONUP:
        if (pressed)
        {
            pressed = false;
            redraw();
            sound::PlaySound("menuok");
            if (listener)
                listener->on_action(this);
        }
        break;
    default:
        break;
    }
}

void 
TextButton::draw(px::GC &gc) 
{
    Button::draw(gc);
    Font *f = pressed ? menufont_pressed : menufont;
    int h = f->get_height();
    int w = f->get_width(text);
    int x = get_area().x + (get_area().w-w)/2;
    int y = get_area().y + (get_area().h-h)/2;
    std::auto_ptr<Surface> s(f->render(text));
    blit(gc, x,y, s.get());
}


//----------------------------------------
// Menu implementation.
//----------------------------------------

void 
Menu::invalidate_area(const px::Rect &r)
{
    dirtyrects.add(r);
}

void 
Menu::invalidate_all()
{
    dirtyrects.clear();
    dirtyrects.push_back(screen->size());
}


void 
Menu::manage(px::Screen *scr)
{
    screen=scr;
    draw_all(scr);
    quit_menu=false;
    while (!quit_menu) {
        screen->flush_updates();
        SDL_Event e;
        while (SDL_PollEvent(&e)) {
            handle_event(e);
        }
        if (!dirtyrects.empty()) {
            RectList::iterator i = dirtyrects.begin();
            RectList::iterator end = dirtyrects.end();
            for (; i!=end; ++i)
                redraw_area(*i);
            dirtyrects.clear();
        }
    }
}

void 
Menu::handle_event(const SDL_Event &e)
{
    on_event(e);
    switch (e.type) {
    case SDL_QUIT:
        quit_menu=true;
        break;
    case SDL_MOUSEMOTION:
        track_active_widget( e.motion.x, e.motion.y );
        break;
    case SDL_KEYDOWN:
        if (e.key.keysym.sym == SDLK_ESCAPE) {
            quit_menu=true;
        }
        else if (active_widget) {
            active_widget->on_event(e);
        }
        break;
    case SDL_MOUSEBUTTONDOWN:
    case SDL_MOUSEBUTTONUP:
    default:
        track_active_widget( e.button.x, e.button.y );
        if (active_widget) active_widget->on_event(e);
        break;
    }
}

void Menu::track_active_widget( int x, int y ) {
    Widget *w = find_widget(x,y);
    if (w != active_widget) 
    {
        if (active_widget) active_widget->deactivate();
        if (w) w->activate();
        active_widget = w;
    }
}


void 
Menu::center(px::Screen *scr)
{
    if (widgets.size() > 0) 
    {
        using std::min;
        using std::max;

        Rect a = widgets[0]->get_area();
        for (unsigned i=1; i<widgets.size(); ++i)
        {
            Rect r = widgets[i]->get_area();
            a.x = min(r.x, a.x);
            a.y = min(r.y, a.y);
            a.w += max(0, r.x+r.w-a.x-a.w);
            a.h += max(0, r.y+r.h-a.y-a.h);
        }
        Rect c=px::center(scr->size(), a);
        int dx = c.x-a.x;
        int dy = c.y-a.y;

        for (unsigned i=0; i<widgets.size(); ++i)
        {
            Rect r = widgets[i]->get_area();
            r.x += dx;
            r.y += dy;
            widgets[i]->set_area(r);
        }
    }    
}


void 
Menu::redraw_area(const px::Rect &r)
{
    video::HideMouse();
    GC gc(screen);
    clip(gc, r);
    draw_background(gc);
    for (unsigned i=0; i<widgets.size(); ++i)
    {
        widgets[i]->draw(gc);
    }
    screen->update_rect(r);
    video::ShowMouse();
}

void 
Menu::draw_all(px::Screen *scr)
{
    redraw_area(scr->size());
    scr->update_all();
}

Widget *
Menu::find_widget(int x, int y)
{
    for (unsigned i=0; i<widgets.size(); ++i)
        if (widgets[i]->get_area().contains(x,y))
            return widgets[i];
    return 0;
}
