//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#include "enigma.hh"
#include "objects.hh"
#include "object_mixins.hh"
#include "display.hh"
#include "player.hh"
#include "lua.hh"
#include "sound.hh"
#include "laser.hh"
#include "actors.hh"

#include "px/buffer.hh"
#include "px/tools.hh"

#include <algorithm>
#include <map>
#include <string>
#include <cstdlib>
#include <iostream>
#include <cassert>

using namespace std;
using namespace world;
using namespace enigma;

using tools::Value;

namespace
{
    player::Inventory *
    get_inventory(const Actor *a)
    {
        if (const Value *v = a->get_attrib("player"))
            return player::GetInventory(to_int(*v));
        return 0;
    }

    bool wielded_item_is(Actor *a, const string &kind)
    {
        if (player::Inventory *inv = get_inventory(a)) 
            if (Item *it = inv->get_item(0))
                return it->get_kind() == kind;
        return false;
    }

}

// This function is used by all triggers, switches etc. that perform
// some particular action when activated (like opening doors or
// switching lasers on and off).
void 
world::PerformAction(Object *o, bool onoff)
{
    try {
        string action = o->string_attrib("action");
        string target = o->string_attrib("target");
        if (action == "callback") {
            lua::CallFunc(target, Value(onoff));
        }
        else if (action == "idle") {
        }
        else if (Object *o = GetNamedObject(target)) {
            SendMessage(o, action);
        }
    } 
    catch (UndefinedAttrib &e) {
        cerr << "perform_action: undefined attrib " << endl;
    }
}

//----------------------------------------
// Object impl.
//----------------------------------------

Object::Object(const string& kind)
{ 
    kind_ = kind;
    set_attrib("kind", Value(kind.c_str()));
}

void 
Object::set_attrib(const string& key, const Value& val)
{
    attribs.assign(key, val);
}

const Value*
Object::get_attrib(const string& key) const
{
    try {
        return &attribs.value(key);
    }
    catch (tools::UnboundName) {
        return 0;
    }
}

const char *
Object::string_attrib(const string &name) const
{
    if (const Value *v = get_attrib(name)) {
        const char *s = to_string(*v);
        if (s == 0)
            throw UndefinedAttrib();
        return s;
    } else
        throw UndefinedAttrib();
}

int 
Object::int_attrib(const string &name) const
{
    if (const Value *v = get_attrib(name))
        return to_int(*v);
    else
        throw UndefinedAttrib();
}

double
Object::double_attrib(const string &name) const
{
    if (const Value *v = get_attrib(name))
        return to_double(*v);
    else
        throw UndefinedAttrib();
}

//----------------------------------------
// GridObject
//----------------------------------------
void 
GridObject::creation(GridPos p) 
{
    pos = p;
    laser::MaybeRecalcLight(p);
    on_creation();
}


//======================================================================
// FLOORS
//======================================================================

Floor::Floor(const string &kind, double friction, double mfactor)
    : GridObject(kind) , _friction(friction)
{
    _mousefactor = mfactor;
}

display::Model*
Floor::set_model(const string & mname)
{
    return display::SetModel(GridLoc(GRID_FLOOR, get_pos()), mname);
}

display::Model *
Floor::get_model() {
    return display::GetModel(GridLoc(GRID_FLOOR, get_pos()));
}

void Floor::on_creation() 
{
    set_model(get_kind());
}

void Floor::on_removal()
{
    display::KillModel(GridLoc(GRID_FLOOR, get_pos()));
}

Object *Floor::clone() {
    return new Floor(get_kind(), _friction, _mousefactor);
}

void Floor::dispose() {
    delete this;
}

void 
Floor::message(const string& msg, const Value &val)
{
    if (msg == "explode")
        SetFloor(get_pos(), MakeFloor("fl-abyss"));
}

//----------------------------------------
// Abyss
//----------------------------------------
namespace
{
    class Abyss : public CloneFloor<Abyss> {
    public:
        Abyss() : Floor("fl-abyss", 10, 1) {}
    private:
        void actor_enter(Actor* a) {SendMessage(a, "fall");}
        void on_actorhit(Actor* a) {SendMessage(a, "fall");}
    };
}

//----------------------------------------
// Water
//----------------------------------------
namespace
{
    class Water : public CloneFloor<Water> {
    public:
        Water() : Floor("fl-water", 5, 1) {}
    private:
        void actor_enter(Actor *a) {SendMessage(a, "drown");}
        void on_actorhit(Actor *a) {SendMessage(a, "drown");}
    };
}

//----------------------------------------
// Gradient
//----------------------------------------

/** \page fl-gradient Gradient Floor

This is a sloped floor that accelerates in a particular direction. 

\subsection gradienta Attributes

- \b type: number between 0 and 15 (see below)

\subsection gradiente Examples

*/
namespace
{
    class Gradient : public CloneFloor<Gradient> {
    public:
        Gradient();
    private:
        int get_type() const;
        enum { MINTYPE=1, MAXTYPE=24 };
        
        // GridObject interface.
        virtual void on_creation();

        // Floor interface
        virtual px::V3 get_force(Actor *a);
    };
}

Gradient::Gradient()
    : Floor("fl-gradient", 12, 2)
{
    set_attrib("type", Value(MINTYPE));
}

int Gradient::get_type() const 
{ 
    int t=int_attrib("type");
    if (t < MINTYPE || t>MAXTYPE) {
        fprintf(stderr, "gradient:: Illegal type=%d\n", t);
        t=MINTYPE;
    }
    return t;
}

void Gradient::on_creation()
{
    char mname[20];
    sprintf(mname, "fl-gradient%d", get_type());
    set_model(mname);
}

px::V3 
Gradient::get_force(Actor *a) 
{
    V3 force;
    int t = get_type();

    static int xforce[MAXTYPE-MINTYPE+1] = {
        0, 0, 1, -1,
        1, -1, 1, -1,
        1, 1, -1, -1,
        -1, -1, 1, 1,
        1, -1, 1, -1,
        0, 0, 1, -1
    };
    static int yforce[MAXTYPE-MINTYPE+1] = {
        1, -1, 0, 0,
        1, 1, -1, -1, 
        1, -1, 1, -1,
        1, -1, 1, -1,
        -1, -1, 1, 1,
        1, -1, 0, 0
    };
    force = V3(xforce[t-MINTYPE], yforce[t-MINTYPE], 0);

    return 20*force;
}


//----------------------------------------
// Bridge
//----------------------------------------

/** \page fl-bridge Bridge Floor

This Floor can be opened and closed much like a bridge.
The actor can move over this floor if the bridge is closed,
and will fall into abyss when the bridge is open.

\subsection bridgea Attributes

- \b type	The type of the bridge, currently only 'a' is possible

\subsection bridgem Messages
- \b open	open the bridge so actors cannot pass it
- \b close	close the bridge so actors can pass it
- \b openclose	toggle the state of the bridge

*/
namespace
{
    class Bridge : public CloneFloor<Bridge> {
    public:
        Bridge();
    private:
        enum State { OPEN, CLOSED, OPENING, CLOSING } state;
	
        string get_type() const {return string_attrib("type");}

        void actor_enter(Actor *);
        void message(const string &m, const Value &);
        void set_model();

        void on_creation() {set_model();}

        void change_state( State newstate);
        void animcb();
    };
}

Bridge::Bridge() : Floor("fl-bridge", 5, 1) 
{
    set_attrib("type", "a");
    state=OPEN;
}

void Bridge::message(const string &m, const Value &) 
{
    if (m == "open" && (state==CLOSED || state==CLOSING))
        change_state(OPENING);
    else if (m=="close" && (state==OPEN || state==OPENING))
        change_state(CLOSING);
    else if (m=="openclose") {
        if (state==OPEN || state==OPENING)
            change_state(CLOSING);
        else
            change_state(OPENING);
    }
}

void Bridge::actor_enter(Actor *a)
{
    if (state != CLOSED)
        SendMessage(a, "fall");
}

void Bridge::set_model() 
{
    string mname = string("fl-bridge")+get_type();	
    if( state==OPEN)
        Floor::set_model(mname+"-open");
    else
        Floor::set_model(mname+"-closed");
}

void Bridge::change_state( State newstate) 
{
    string mname = string("fl-bridge")+get_type();
    switch( newstate) {
    case OPENING:
        if( state == CLOSING)
            get_model()->reverse();
        else
            Floor::set_model(mname+"-opening")
                ->set_callback(px::make_cb(this, &Bridge::animcb));
        break;
    case CLOSING:
        if( state == OPENING)
            get_model()->reverse();
        else
            Floor::set_model(mname+"-closing")
                ->set_callback(px::make_cb(this, &Bridge::animcb));
        break;
    case OPEN:
    case CLOSED:
        state=newstate;
        set_model();
        break;
    }
    state=newstate;
}

void Bridge::animcb() 
{
    if (state==OPENING)
        change_state(OPEN);
    if (state==CLOSING)
        change_state(CLOSED);
}

#if 0

class WhiteTile : public Floor {
public:
    V3 get_force(Actor *a, V3 mouseforce) {
        if (a->get_attrib("blackball"))
            return Floor::get_force(a, 0);
        else
            return Floor::get_force(a, mouseforce);
    }
};

class BlackTile : public Floor {
public:
    V3 get_force(Actor *a, V3 mouseforce) {
        if (a->get_attrib("whiteball"))
            return Floor::get_force(a, 0);
        else
            return Floor::get_force(a, mouseforce);
    }
};

#endif


//======================================================================
// STONES
//======================================================================


Stone::Stone(const string & kind) 
    : GridObject(kind) 
{}

void Stone::on_creation() {
    set_model();
}

void Stone::set_model() {
    set_model(get_kind());
}

display::Model*
Stone::set_model(const string & mname)
{
    return display::SetModel(GridLoc(GRID_STONES, get_pos()), mname);
}

display::Model *
Stone::get_model() {
    return display::GetModel(GridLoc(GRID_STONES, get_pos()));
}

void 
Stone::on_removal()
{
    display::KillModel(GridLoc(GRID_STONES, get_pos()));
    laser::MaybeRecalcLight(get_pos());
}

px::V3 
Stone::actor_impulse(const StoneContact &sc)
{
    V3 normal = normalize(sc.normal);
    const V3 &v = sc.actor->get_vel();
    return -2*(v*normal)*normal;
}


StoneResponse 
Stone::actor_hit(const StoneContact &sc)
{
    return STONE_REBOUND;
}

const char *
Stone::collision_sound()
{
    return "st-stone";
}

//----------------------------------------
// SimpleStone
//
// This kind of stone can be defined from Lua programs with
// `DefineSimpleStone'.
//----------------------------------------
namespace
{
    class SimpleStone : public Stone {
    public:
        SimpleStone(const string &knd, const string & snd)
            : Stone(knd.c_str()), sound(snd)
        {}
    private:
        const char *collision_sound() {
            return sound.c_str();
        }

        Object *clone() {
            return new SimpleStone(get_kind(), sound);
        }
        void dispose() { delete this; }

        string sound;
    };
}

//----------------------------------------
// One Way Stone
//
// This stone can only be passed in one direction. 
//----------------------------------------
namespace
{
    class OneWayStone : public CloneStone<OneWayStone> {
    public:
	OneWayStone() : Stone("st-oneway") {
	    set_orientation(SOUTH);
	}
    private:
	void set_model() {
	    Stone::set_model(string("st-oneway") 
			     + to_suffix(get_orientation()));
	}
	Direction get_orientation() const {
	    return Direction(int_attrib("orientation"));
	}
        void set_orientation(Direction dir) {
            set_attrib("orientation", Value(dir));
        }
	StoneResponse actor_hit(const StoneContact&);
    };
}

StoneResponse
OneWayStone::actor_hit(const StoneContact &sc)
{
    DirectionBits dirs=contact_faces(sc);
    Direction o=get_orientation();
    
    if (has_dir(dirs,o)) {
        // magic wand as first item?
        if (wielded_item_is(sc.actor, "it-magicwand")) {
            set_orientation(reverse(o));
            set_model();
        }
        return STONE_REBOUND;
    } 
    return STONE_PASS;
}

//----------------------------------------
// Chameleon Stone
//
// This stone takes on the look of the floor beneath it.  Actors can
// move through it, so these stones are perfect for hiding stuff under
// them...
//----------------------------------------
namespace
{
    class ChameleonStone : public CloneStone<ChameleonStone> {
    public:
        ChameleonStone() : Stone("st-chameleon"){}
    private:
        void set_model() {
            string modelname = "fl-gray";
            if (Floor *fl = GetFloor(get_pos()))
                modelname = fl->get_kind();
            Stone::set_model(modelname);
        }
        StoneResponse actor_hit(const StoneContact &sc) {
            return STONE_PASS;
        }
    };
}


//----------------------------------------
// Key Switch, Key_a Switch, Key_b Switch
//
// Attributes:
//
// :keycode         a numerical code; only keys with the same code 
//                  can activate this switch
// :on              1 or 0
// :target,action   as usual
//----------------------------------------
namespace
{
    class KeyStone : public OnOffStone
    {
    protected:
        KeyStone();
    private:
        void set_model() {
            Stone::set_model(is_on() ? "st-key1" : "st-key0");
        }
        StoneResponse actor_hit(const StoneContact &sc);
        const char *collision_sound() { return "st-metal"; }
    };

    class KeyStone_a : public KeyStone, public CloneStone<KeyStone_a>  {
    public:
        KeyStone_a() : Stone("st-key_a") {
            set_attrib("keycode", 1.0);
        }
    };
    class KeyStone_b : public KeyStone, public CloneStone<KeyStone_b> {
    public:
        KeyStone_b() : Stone("st-key_b") {
            set_attrib("keycode", 2.0);
        }
    };
    class KeyStone_c : public KeyStone, public CloneStone<KeyStone_c> {
    public:
        KeyStone_c() : Stone("st-key_c") {
            set_attrib("keycode", 3.0);
        }
    };
}

KeyStone::KeyStone() : Stone("st-key") 
{
    set_attrib("keycode", 0.0);
}

StoneResponse 
KeyStone::actor_hit(const StoneContact &sc) 
{
    if (player::Inventory *inv = get_inventory(sc.actor))
    {
        if (is_on()) {
            Item *key = MakeItem("it-key");
            key->set_attrib("keycode", int_attrib("keycode"));
            inv->add_item(key);
            set_on(false);
            PerformAction(this, is_on());
        }
        else // if (wielded_item_is(sc.actor, "it-key")) 
        {
            try {
                Item *it = inv->get_item(0);
                if (it && it->int_attrib("keycode") == int_attrib("keycode"))
                {
                    set_on(true);
                    PerformAction(this, is_on());
                    delete inv->yield_first();
                }
            }
            catch (UndefinedAttrib)
            {}
        }
    }
    return STONE_REBOUND;
}

//----------------------------------------
// SwapStone
//----------------------------------------

/** \page st-swap Swap Stone

This stone can exchange its position with other neighboring stones
if it is hit hard enough.  In a way, this makes swap stones a kind
of "movable stone", except that they can be only exchanged with
other stones and may not be moved on empty fields.

\subsection swape Example
\verbatim
set_stone("st-swap", 10,10)
\endverbatim

\image html st-swap.png
*/
namespace
{
    class SwapStone : public CloneStone<SwapStone> {
    public:
        SwapStone() : Stone("st-swap") {}
    private:
        StoneResponse actor_hit(const StoneContact &sc)
        {
            Direction hit_dir = reverse(contact_face(sc));
            Actor *a = sc.actor;
            if (a->get_vel() * sc.normal < -4 && hit_dir!=NODIR) {
                GridPos p = get_pos();
                GridPos newp = move(p, hit_dir);
                if (GetStone(newp))
                    SwapStones(p,newp);
            }
            return STONE_REBOUND;
        }
    };
}

//----------------------------------------
// Greenbrown_move
//----------------------------------------
namespace
{
    class Greenbrown_move : public CloneStone<Greenbrown_move>, 
                            public MovableStone 
    {
        const char *collision_sound() {return "st-thud";}
    public:
        Greenbrown_move() : Stone("st-greenbrown_move") {}
    };
}

//----------------------------------------
// Marble_move
//----------------------------------------
namespace
{
    class Marble_move : public CloneStone<Marble_move>,
                        public MovableStone
    {
        const char *collision_sound() {return "st-thud";}
    public:
        Marble_move() : Stone("st-marble_move") {}
    };
}

//----------------------------------------
// Rock3_move
//----------------------------------------
namespace
{
    class Rock3_move : public CloneStone<Rock3_move>,
                       public MovableStone
    {
        const char *collision_sound() {return "st-thud";}
    public:
        Rock3_move() : Stone("st-rock3_move") {}
    };
}

//----------------------------------------
// Stone_break
//----------------------------------------
namespace
{
    class Stone_break : public CloneStone<Stone_break> {
        const char *collision_sound() {return "st-thud";}
    public:
        Stone_break() : Stone("st-stone_break") {}

        StoneResponse actor_hit(const StoneContact &sc) {
            if (wielded_item_is(sc.actor, "it-hammer")) {
                return STONE_KILL;
            }
            return STONE_REBOUND;
       }

    };
}

//----------------------------------------
// Brownie
//----------------------------------------
namespace
{
    class Brownie : public CloneStone<Brownie>, public MovableStone {
        const char *collision_sound() {return "st-thud";}
    public:
        Brownie() : Stone("st-brownie") {}
    };
}

//----------------------------------------
// Brick Magic
//----------------------------------------
namespace
{
    class BrickMagic : public CloneStone<BrickMagic> {
        const char *collision_sound() {return "st-swap";}
    public:
	BrickMagic() : Stone("st-brick_magic"), state(BRICK) {}
    private:
        enum State { BRICK, GLASS } state;
	StoneResponse actor_hit(const StoneContact &sc)
        {
	    if( state == BRICK)
            {
	        if (wielded_item_is(sc.actor, "it-magicwand")) {
		    state = GLASS;
		    Stone::set_model("st-glass");
		    laser::MaybeRecalcLight(get_pos());
                }
            }
	    return STONE_REBOUND;
	}
	bool on_laserhit(Direction dir)
	{
	    if( state == BRICK)
                return false;
            return true;
	}
    };
}

//----------------------------------------
// WoodenStone
//----------------------------------------
namespace
{
    class WoodenStone : public CloneStone<WoodenStone>, public MovableStone {
    public:
        WoodenStone() : Stone("st-wood") {}
    private:

        StoneResponse actor_hit(const StoneContact &sc)
        {
            MovableStone::actor_hit(sc);
            if (Floor *fl=GetFloor(get_pos())) {
                const string &k = fl->get_kind();
                if (k == "fl-abyss" || k=="fl-water" || k=="fl-swamp") {
                    SetFloor(get_pos(), MakeFloor("fl-wood"));
                    return STONE_KILL;
                }
            }
            return STONE_REBOUND;
        }
    };
}

//----------------------------------------
// ElectricStone
//
// Attributes:
//
// :charge  + - 0
//----------------------------------------
namespace
{
    class ElectricStone : public CloneStone<ElectricStone> {
    public:
        ElectricStone() : Stone("st-electric") {
            set_attrib("charge", "+");
        }
    private:
        void set_model() {
            string c = string_attrib("charge");
            if (c=="0")
                ;
            else if (c=="+")
                ;
            else if (c=="-")
                ;
        }
    };
}

//----------------------------------------
// ScissorsStone
//----------------------------------------
namespace
{
    class ScissorsStone : public CloneStone<ScissorsStone> {
    public:
        ScissorsStone() : Stone("st-scissors") {}
    private:
        StoneResponse actor_hit(const StoneContact &sc) {
            // cut all rubber bands connected to sc.actor
            return STONE_REBOUND;
        }
    };
}

//----------------------------------------
// BolderStone
//----------------------------------------

/** \page st-bolder Bolder Stone

\subsection boldera Attributes

- \c direction  \n NORTH, EAST, SOUTH, WEST

*/
namespace
{
    class BolderStone 
        : public CloneStone<BolderStone>,
          public tools::TimeHandler 
    {
    public:
        BolderStone() : Stone("st-bolder"), state(IDLE) {
            set_attrib("direction", enigma::NORTH);
        }
    private:
        enum State { IDLE, MOVING } state;

        Direction get_dir() const 
        { 
            return static_cast<Direction>(int_attrib("direction")); 
        }
        void set_dir(Direction d) {set_attrib("direction", d);}

        bool on_move() { return true; }

        void on_creation() {
            if (state == IDLE)
                g_timer.set_alarm(this, 0.3, true);
            Stone::on_creation();
        }
        void on_removal() {
            if (state == IDLE)
                g_timer.remove_alarm(this);
            Stone::on_removal();
        }

        void alarm() {
            state = MOVING;
	    GridPos p = get_pos();
            if (world::MaybeMoveStone(p, get_dir())) {
                // send a message to the stone that's blocking our way
		Stone *st = GetStone(move(move(p,get_dir()),get_dir()));
		if( st)
	            SendMessage(st, "trigger");
            }
            state = IDLE;
        }

        // Stone interface.
        void set_model() 
        {
            string mname = "st-bolder" + to_suffix(get_dir());
            Stone::set_model(mname);
        }

        StoneResponse actor_hit(const StoneContact &sc)
        {
            Actor *a = sc.actor;

            if (wielded_item_is(a, "it-magicwand")) {
                set_dir(reverse(get_dir()));
                set_model();
            }
            return STONE_REBOUND;
        }
        void actor_inside(Actor *a) {SendMessage(a, "shatter");}
    };
}


//----------------------------------------
// PuzzleStones
//----------------------------------------

/** \page st-puzzle Puzzle Stone

These stones can be connected to other stones of the same type.  Any
of the four faces of the stone can have ``socket''.  If the touching
faces of two adjacent stones both have a socket, the two stones link
up and henceforth move as group.

A cluster of puzzle stones may for example look like this:

\verbatim
+---+---+---+---+
|   |   |   |   |
| --+-+-+---+-+ |
|   | | |   | | |
+---+-+-+---+-+-+
    | | |   | | |
    | | |   | | |
    |   |   |   |
    +---+   +---+
\endverbatim

This example actually presents the special case of a ``complete''
cluster.  A cluster is complete if none of its stones has an
unconnected socket.  If such a group is touched with a magic wand, all
its constituents explode.  [They should, anyway. This is not
implemented yet.]

\subsection puzzlea Attributes

- \b connections 
   number between 1 an 16.  Each bit in (connections-1) corresponds to
   a socket on one of the four faces.  You will normally simply use
   one of the Lua constants \c PUZ_0000 to \c PUZ_1111.

\subsection puzzlee Example
<table>
<tr>
<td> \image html st-puzzletempl_0001.png "PUZ_0000"
<td> \image html st-puzzletempl_0002.png "PUZ_0001"
<td> \image html st-puzzletempl_0003.png "PUZ_0010"
<td> \image html st-puzzletempl_0004.png "PUZ_0011"
<tr>
<td> \image html st-puzzletempl_0005.png "PUZ_0100"
<td> \image html st-puzzletempl_0006.png "PUZ_0101"
<td> \image html st-puzzletempl_0007.png "PUZ_0110"
<td> \image html st-puzzletempl_0008.png "PUZ_0111"
<tr>
<td> \image html st-puzzletempl_0009.png "PUZ_1000"
<td> \image html st-puzzletempl_0010.png "PUZ_1001"
<td> \image html st-puzzletempl_0011.png "PUZ_1010"
<td> \image html st-puzzletempl_0012.png "PUZ_1011"
<tr>
<td> \image html st-puzzletempl_0013.png "PUZ_1100"
<td> \image html st-puzzletempl_0014.png "PUZ_1101"
<td> \image html st-puzzletempl_0015.png "PUZ_1110"
<td> \image html st-puzzletempl_0016.png "PUZ_1111"
</tr>
</table>
*/
namespace
{
    class PuzzleStone : public Stone {
    public:
        PuzzleStone() : Stone("st-puzzle") {
            set_attrib("connections", 0.0);
        }
    private:
        typedef vector<PuzzleStone*> InstanceList;
        typedef vector<GridPos> Cluster;

        DirectionBits get_connections();

        static bool visit_dir(vector<GridPos> &stack, 
                              GridPos curpos, Direction dir);
        bool find_cluster(Cluster &);
        void find_row_or_column_cluster(Cluster &c, 
                                        GridPos startpos, Direction dir);
        bool cluster_complete();
        void maybe_move_cluster(Cluster &c, Direction dir);
        void rotate_cluster(const Cluster &c);

        bool on_laserhit(Direction dir) { 
            return get_connections() == 0; 
        }

        void set_model();
        StoneResponse actor_hit(const StoneContact &sc);
        void actor_inside(Actor *a) {
            if (get_connections() != NODIRBIT)
                SendMessage(a, "shatter");
        }

        Object *clone() {
            PuzzleStone *o = new PuzzleStone;
            instances.push_back(o);
            return o;
        }
        void dispose() {
            instances.erase(find(instances.begin(), instances.end(), this));
            delete this;
        }

        // Variables
        bool visited;           // flag for DFS
        static InstanceList instances;
    };
}

PuzzleStone::InstanceList PuzzleStone::instances;

DirectionBits 
PuzzleStone::get_connections() 
{
    int conn=int_attrib("connections") - 1;
    if (conn >=0 && conn <16)
        return DirectionBits(conn);
    else
        return NODIRBIT;
}

bool
PuzzleStone::visit_dir(vector<GridPos> &stack, GridPos curpos, Direction dir)
{
    GridPos newpos = move(curpos, dir);
    PuzzleStone *pz = dynamic_cast<PuzzleStone*>(GetStone(newpos));
    
    if (!pz)
        return false;

    DirectionBits cfaces = pz->get_connections();

    if (cfaces==NODIRBIT || has_dir(cfaces, reverse(dir))) {
        // Puzzle stone at newpos is connected to stone at curpos
        if (!pz->visited)
            stack.push_back(newpos);
        return true;
    } else {
        // The two stones are adjacent but not connected
        return false;
    }
}

// Use a depth first search to determine the group of all stones that
// are connected to the current stone.  Returns true if the cluster is
// ``complete'' in the sense defined above.
bool
PuzzleStone::find_cluster(Cluster &cluster)
{
    for (unsigned i=0; i<instances.size(); ++i)
        instances[i]->visited=false;

    vector<GridPos> pos_stack;
    bool is_complete = true;
    pos_stack.push_back(get_pos());
    while (!pos_stack.empty()) {
        GridPos curpos = pos_stack.back();
        pos_stack.pop_back();

        PuzzleStone *pz = dynamic_cast<PuzzleStone*>(GetStone(curpos));
        assert(pz);

        cluster.push_back(curpos);
        pz->visited = true;
        DirectionBits cfaces = pz->get_connections();

        if (cfaces==NODIRBIT) 
            cfaces = DirectionBits(NORTHBIT | SOUTHBIT | EASTBIT | WESTBIT);

        if (has_dir(cfaces, NORTH))
            is_complete &= visit_dir(pos_stack, curpos, NORTH);
        if (has_dir(cfaces, EAST))
            is_complete &= visit_dir(pos_stack, curpos, EAST);
        if (has_dir(cfaces, SOUTH))
            is_complete &= visit_dir(pos_stack, curpos, SOUTH);
        if (has_dir(cfaces, WEST))
            is_complete &= visit_dir(pos_stack, curpos, WEST);
    }
    return is_complete;
}

void
PuzzleStone::find_row_or_column_cluster(Cluster &c, GridPos startpos, Direction dir)
{
    GridPos p = startpos;
    while (dynamic_cast<PuzzleStone*>(GetStone(p))) {
        c.push_back(p);
        p.move(dir);
    }
}

void 
PuzzleStone::maybe_move_cluster(Cluster &c, Direction dir)
{
    // First remove all stones in the cluster from the world
    vector<Stone*> clusterstones;
    for (unsigned i=0; i<c.size(); ++i) {
        clusterstones.push_back(YieldStone(c[i]));
        c[i] = move(c[i], dir);
    }

    // Now check whether all stones can be placed at their new
    // position
    bool move_ok = true;
    for (unsigned i=0; i<c.size(); ++i) {
        if (GetStone(c[i]) != 0)
            move_ok = false;
    }

    // Now either move all the stones or place them at their old
    // positions
    for (unsigned i=0; i<c.size(); ++i) {
        if (move_ok)
            SetStone(c[i], clusterstones[i]);
        else
            SetStone(move(c[i], reverse(dir)), clusterstones[i]);
    }
    if (move_ok) 
        sound::PlaySound("st-move");
}

bool 
PuzzleStone::cluster_complete()
{
    Cluster c;
    return find_cluster(c);
}

void 
PuzzleStone::set_model()
{
    string modelname = "st-puzzle";
    char x[10];

    // every combination of ``sockets'' has its own model
    int modelno = int_attrib("connections");
    sprintf(x, "%d", modelno);
    Stone::set_model(modelname+x);
}

void
PuzzleStone::rotate_cluster(const Cluster &c)
{
    if (c.size() > 1) {
        vector<Stone*> stones;
        for (unsigned i=0; i<c.size(); ++i)
            stones.push_back(YieldStone(c[i]));

        // exclude last list item
        for (unsigned i=0; i<c.size()-1; ++i)
            SetStone(c[i+1], stones[i]);
        SetStone(c[0], stones.back());
    }
}

StoneResponse 
PuzzleStone::actor_hit(const StoneContact &sc)
{
    if (get_connections() == NODIRBIT)
        return STONE_PASS;

    Actor *a = sc.actor;

    bool move_impulse= (a->get_vel()*sc.normal < -4);
    Direction hit_dir=reverse(contact_face(sc));

    if (wielded_item_is(a, "it-magicwand")) {
        if (cluster_complete()) {
            // make all cluster stones explode
            return STONE_REBOUND;
        } 
        else if (hit_dir != NODIR) {
            Cluster c;
            find_row_or_column_cluster(c, get_pos(), hit_dir);

            if (move_impulse)
                maybe_move_cluster(c, hit_dir);
            else
                rotate_cluster(c);
        }
    } 
    else if (move_impulse && hit_dir != NODIR) {
        Cluster c;
        find_cluster(c);
        maybe_move_cluster(c, hit_dir);
    }

    return STONE_REBOUND;
}


//----------------------------------------
// DoorBase
//
// Base class for everything that behaves like a door, i.e., it has
// four states OPEN, CLOSED, OPENING, CLOSING.
//----------------------------------------
namespace
{
    class DoorBase
    {
    protected:
        DoorBase() : state(CLOSED) {}
        enum State { OPEN, CLOSED, OPENING, CLOSING };

        State get_state() const { return state; }
        void set_state(State st) { state=st; }

        State state;

    private:

        void actor_inside(Actor *a) {
            if (get_state() == CLOSING)
                SendMessage(a, "shatter");
        }

    };
}

//----------------------------------------
// Door
//
// Attributes:
//
// :type        h or v for a door that opens horizontally or vertically
//----------------------------------------
namespace
{
    class Door : public CloneStone<Door>, public DoorBase {
    public:
        Door() : Stone("st-door") {
            set_attrib("type", "h");
        }
    protected:
    private:
        string get_type() const {return string_attrib("type");}

        void animcb();
        void set_model();

        bool on_laserhit(Direction dir) {
            if (get_type() == "h")
                return state==OPEN || dir==EAST || dir==WEST;
            else
                return state==OPEN || dir==NORTH || dir==SOUTH;
        }

        void message(const string &m, const Value &);
        void change_state(State newstate);
        StoneResponse actor_hit(const StoneContact &sc);
        void actor_inside(Actor *a);
        void actor_contact(Actor *a);
    };
}

void 
Door::set_model() 
{
    string mname = string("st-door")+get_type();
    if (state == CLOSED) 
        mname += "-closed";
    else if (state==OPEN)
        mname += "-open";
    Stone::set_model(mname);
}

void 
Door::animcb()
{
    if (state == OPENING)
        change_state(OPEN);
    else if (state == CLOSING)
        change_state(CLOSED);
}

void 
Door::message(const string &m, const Value &)
{
    if (m == "open" && (state==CLOSED || state==CLOSING)) 
        change_state(OPENING);
    else if (m=="close" && (state==OPEN || state==OPENING))
        change_state(CLOSING);
    else if (m=="openclose") {
        if (state==OPEN || state==OPENING)
            change_state(CLOSING);
        else 
            change_state(OPENING);
    }
}

void 
Door::change_state(State newstate) 
{
    string basename = string("st-door") + get_type();
    switch (newstate) {
    case OPEN: 
        Stone::set_model(basename+"-open"); 
        laser::MaybeRecalcLight(get_pos());
        break;
    case CLOSED: 
        Stone::set_model(basename+"-closed"); 
        laser::MaybeRecalcLight(get_pos());
        break;
    case OPENING:
//         sound::PlaySound("dooropen");
        if (state == CLOSING)
            get_model()->reverse();
        else {
            Stone::set_model(basename+"-opening")
                ->set_callback(px::make_cb(this, &Door::animcb));
        }
        break;
    case CLOSING:
//         sound::PlaySound("doorclose");
        if (state == OPENING)
            get_model()->reverse();
        else {
            Stone::set_model(basename+"-closing")
                ->set_callback(px::make_cb(this, &Door::animcb));
        }
        break;
    }
    set_state(newstate);
}


StoneResponse 
Door::actor_hit(const StoneContact &sc)
{
    Direction cf = contact_face(sc);
    if (state == OPEN)
        return STONE_PASS;
    else if (state == CLOSING) {
        SendMessage(sc.actor, "shatter");
        return STONE_REBOUND;
    } 
    else {
        string t = get_type();
        return ((t == "v" && (cf==WEST || cf==EAST)) ||
                (t == "h" && (cf==SOUTH || cf==NORTH)))
            ? STONE_REBOUND 
            : STONE_PASS;
    }
}

void 
Door::actor_inside(Actor *a)
{
    if (state == CLOSING)
        SendMessage(a, "shatter");
}

void 
Door::actor_contact(Actor *a)
{
    if (state == CLOSING)
        SendMessage(a, "shatter");
}


//----------------------------------------
// Door_a
//----------------------------------------
namespace
{
    class Door_a : public CloneStone<Door_a>, public DoorBase {
    public:
        Door_a() : Stone("st-door_a") {}
    private:
        void animcb();

        void set_model();

        bool on_laserhit(Direction dir) {
            return get_state()==OPEN;
        }

        void message(const string &m, const Value &);
        void change_state(State newstate);
        StoneResponse actor_hit(const StoneContact &sc);
    };
}

void
Door_a::set_model() {
    string mname = string("st-doora");
    if (state == CLOSED)
        mname += "-closed";
    else if (state==OPEN)
        mname += "-open";
    Stone::set_model(mname);
}

void
Door_a::animcb()
{
    if (state == OPENING)
        change_state(OPEN);
    else if (state == CLOSING)
        change_state(CLOSED);
}

void
Door_a::message(const string &m, const Value &)
{
    if (m == "open" && (state==CLOSED || state==CLOSING))
        change_state(OPENING);
    else if (m=="close" && (state==OPEN || state==OPENING))
        change_state(CLOSING);
    else if (m=="openclose") {
        if (state==OPEN || state==OPENING)
            change_state(CLOSING);
        else
            change_state(OPENING);
    }
}

void
Door_a::change_state(State newstate)
{
    string basename = string("st-doora");
    switch (newstate) {
    case OPEN:
        Stone::set_model(basename+"-open");
        laser::MaybeRecalcLight(get_pos());
        break;
    case CLOSED:
        Stone::set_model(basename+"-closed");
        laser::MaybeRecalcLight(get_pos());
        break;
    case OPENING:
//         sound::PlaySound("door_aopen");
        if (state == CLOSING)
            get_model()->reverse();
        else {
            Stone::set_model(basename+"-opening")
                ->set_callback(px::make_cb(this, &Door_a::animcb));
        }
        break;
    case CLOSING:
//         sound::PlaySound("door_aclose");
        if (state == OPENING)
            get_model()->reverse();
        else {
            Stone::set_model(basename+"-closing")
                ->set_callback(px::make_cb(this, &Door_a::animcb));
        }
        break;
    }
    state = newstate;
}


StoneResponse
Door_a::actor_hit(const StoneContact &sc)
{
    if (state == OPEN)
        return STONE_PASS;
    else if (state == CLOSING) {
        SendMessage(sc.actor, "shatter");
        return STONE_REBOUND;
    }
    return STONE_REBOUND;
}


//----------------------------------------
// Door_b
//----------------------------------------
namespace
{
    class Door_b : public CloneStone<Door_b> {
    public:
        Door_b() : Stone("st-door_b") {
            set_attrib("type", "h");
            state = CLOSED;
        }
    private:
        enum State { OPEN, CLOSED, OPENING, CLOSING };
        State state;

        string get_type() const {
            return string_attrib("type");
        }

        void animcb();

        void set_model();

        bool on_laserhit(Direction dir) {
            if (get_type() == "h")
                return state==OPEN || dir==EAST || dir==WEST;
            else
                return state==OPEN || dir==NORTH || dir==SOUTH;
        }
        void actor_inside(Actor *a) {
            if (state == CLOSING)
                SendMessage(a, "shatter");
        }

        void message(const string &m, const Value &);
        void change_state(State newstate);
        StoneResponse actor_hit(const StoneContact &sc);
    };
}

void
Door_b::set_model() {
    string mname = string("st-doorb");
    if (state == CLOSED)
        mname += "-closed";
    else if (state==OPEN)
        mname += "-open";
    Stone::set_model(mname);
}

void
Door_b::animcb()
{
    if (state == OPENING)
        change_state(OPEN);
    else if (state == CLOSING)
        change_state(CLOSED);
}

void
Door_b::message(const string &m, const Value &)
{
    if (m == "open" && (state==CLOSED || state==CLOSING))
        change_state(OPENING);
    else if (m=="close" && (state==OPEN || state==OPENING))
        change_state(CLOSING);
    else if (m=="openclose") {
        if (state==OPEN || state==OPENING)
            change_state(CLOSING);
        else
            change_state(OPENING);
    }
}

void
Door_b::change_state(State newstate)
{
    string basename = string("st-doorb");
    switch (newstate) {
    case OPEN:
        Stone::set_model(basename+"-open");
        laser::MaybeRecalcLight(get_pos());
        break;
    case CLOSED:
        Stone::set_model(basename+"-closed");
        laser::MaybeRecalcLight(get_pos());
        break;
    case OPENING:
//         sound::PlaySound("door_bopen");
        if (state == CLOSING)
            get_model()->reverse();
        else {
            Stone::set_model(basename+"-opening")
                ->set_callback(px::make_cb(this, &Door_b::animcb));
        }
        break;
    case CLOSING:
//         sound::PlaySound("door_bclose");
        if (state == OPENING)
            get_model()->reverse();
        else {
            Stone::set_model(basename+"-closing")
                ->set_callback(px::make_cb(this, &Door_b::animcb));
        }
        break;
    }
    state = newstate;
}


StoneResponse
Door_b::actor_hit(const StoneContact &sc)
{
    Direction cf = contact_face(sc);
    if (state == OPEN)
        return STONE_PASS;
    else if (state == CLOSING) {
        SendMessage(sc.actor, "shatter");
        return STONE_REBOUND;
    }
    else {
        string t = get_type();
        return ((t == "v" && (cf==WEST || cf==EAST)) ||
                (t == "h" && (cf==SOUTH || cf==NORTH)))
            ? STONE_REBOUND
            : STONE_PASS;
    }
}

//----------------------------------------
// Door_c
//----------------------------------------

namespace
{
    class Door_c : public CloneStone<Door_c>, public DoorBase {
    public:
        Door_c() : Stone("st-door_c") {}
    private:
        void animcb();

        void set_model();

        bool on_laserhit(Direction dir) {
            return get_state()==OPEN;
        }

        void message(const string &m, const Value &);
        void change_state(State newstate);
        StoneResponse actor_hit(const StoneContact &sc);
    };
}

void
Door_c::set_model() {
    string mname = string("st-doorc");
    if (state == CLOSED)
        mname += "-closed";
    else if (state==OPEN)
        mname += "-open";
    Stone::set_model(mname);
}

void
Door_c::animcb()
{
    if (state == OPENING)
        change_state(OPEN);
    else if (state == CLOSING)
        change_state(CLOSED);
}

void
Door_c::message(const string &m, const Value &)
{
    if (m == "open" && (state==CLOSED || state==CLOSING))
        change_state(OPENING);
    else if (m=="close" && (state==OPEN || state==OPENING))
        change_state(CLOSING);
    else if (m=="openclose") {
        if (state==OPEN || state==OPENING)
            change_state(CLOSING);
        else
            change_state(OPENING);
    }
}

void
Door_c::change_state(State newstate)
{
    string basename = string("st-doorc");
    switch (newstate) {
    case OPEN:
        Stone::set_model(basename+"-open");
        laser::MaybeRecalcLight(get_pos());
        break;
    case CLOSED:
        Stone::set_model(basename+"-closed");
        laser::MaybeRecalcLight(get_pos());
        break;
    case OPENING:
//         sound::PlaySound("door_copen");
        if (state == CLOSING)
            get_model()->reverse();
        else {
            Stone::set_model(basename+"-opening")
                ->set_callback(px::make_cb(this, &Door_c::animcb));
        }
        break;
    case CLOSING:
//         sound::PlaySound("door_cclose");
        if (state == OPENING)
            get_model()->reverse();
        else {
            Stone::set_model(basename+"-closing")
                ->set_callback(px::make_cb(this, &Door_c::animcb));
        }
        break;
    }
    state = newstate;
}


StoneResponse
Door_c::actor_hit(const StoneContact &sc)
{
    if (state == OPEN)
        return STONE_PASS;
    else if (state == CLOSING) {
        SendMessage(sc.actor, "shatter");
        return STONE_REBOUND;
    }
    return STONE_REBOUND;
}

//----------------------------------------
// TimerStone
//
// Attributes:
//
// :interval        seconds between two "ticks"
// :loop            
// :action,target   as usual
//----------------------------------------

/** \page st-timer Timer Stone

This stone can be used to trigger periodic events or to trigger one
single event after a certain amount of time.

\subsection timera Attributes

- \b on: 1 if the timer is running
- \b interval:  number of seconds before \b action is performed
- \b loop:      if 1, restart the timer after performing \b action
- \b action, \b target: as usual

\subsection timerm Messages

- \b on, \b off, \b onoff: as usual

\subsection timere Example

\verbatim
-- activate a laser after 5 seconds
set_stone("st-laser", 10,11, {name="laser"})
set_stone("st-timer", 10,10, 
          {loop=0, action="onoff", target="laser", interval=5})
\endverbatim
*/
namespace
{
    class TimerStone 
        : public CloneStone<TimerStone>,
          public OnOffStone,
          public tools::TimeHandler 
    {
    public:
        TimerStone() : Stone("st-timer") 
        {
            set_attrib("interval", 1.0);
            set_attrib("loop", 1.0);
            set_attrib("on", 1.0);

            // set_on(true);   DOESN'T WORK! calls set_model()
        }
    private:
        double get_interval() const { return double_attrib("interval"); }

        void set_model() {
            if (is_on())
                Stone::set_model("st-timer");
            else
                Stone::set_model("st-timeroff");
        }

        void on_creation() {
            set_alarm();
            Stone::on_creation();
        }

        void set_alarm() {
            if (is_on())
                g_timer.set_alarm(this, get_interval(), true);
        }

        void alarm() {
            if (is_on())
            {
//                 sound::PlaySound("st-timer");
                PerformAction(this, is_on());
            }
        }
    };
}

//----------------------------------------
// Switch
//
// Attributes:
//
// :on              1 or 0
// :target,action   as usual
//----------------------------------------
namespace
{
    class SwitchStone : public CloneStone<SwitchStone>, public OnOffStone {
    public:
        SwitchStone() : Stone("st-switch") {}
    private:
        void set_model() {
            Stone::set_model(is_on() ? "st-switch1" : "st-switch0");
        }

        StoneResponse actor_hit(const StoneContact &sc) {
            set_on(!is_on());
            PerformAction(this, is_on());
            sound::PlaySound("st-switch");
            return STONE_REBOUND;
        }
        const char *collision_sound() { return "st-metal"; }
    };
}

//----------------------------------------
// Floppy Switch
//
// Attributes:
//
// :on              1 or 0
// :target,action   as usual
//----------------------------------------
namespace
{
    class FloppyStone : 
        public CloneStone<FloppyStone>, public OnOffStone 
    {
    public:
        FloppyStone() : Stone("st-floppy") {}
    private:
        void set_model() {
            Stone::set_model(is_on() ? "st-floppy1" : "st-floppy0");
        }

        StoneResponse actor_hit(const StoneContact &sc) 
        {
            if (player::Inventory *inv = get_inventory(sc.actor))
            {
                if (is_on()) 
                {
                    inv->add_item(MakeItem("it-floppy"));
                    set_on(false);
                    PerformAction(this, is_on());
                }
                else if (wielded_item_is(sc.actor, "it-floppy")) 
                {
                    DisposeObject(inv->yield_first());
                    set_on(true);
                    PerformAction(this, is_on());
                }
            }
            return STONE_REBOUND;
        }
        const char *collision_sound() { return "st-metal"; }
    };
}

//----------------------------------------
// ShogunStone
//
// Attributes:
//
// :holes            1..7
//----------------------------------------
namespace
{
    class ShogunStone : public CloneStone<ShogunStone> {
    public:
        ShogunStone() : Stone("st-shogun") {
            set_holes(SMALL);
        }
    private:
        enum Holes { SMALL=1, MEDIUM=2, LARGE=4 };

        void set_holes(Holes h) { set_attrib("holes", h); }
        Holes get_holes() const;
        void notify_item();

        void add_hole(Holes h) {
            set_attrib("holes", get_holes() | h);
            notify_item();
            set_model();
        }

        void on_creation() { 
            set_model(); 
            notify_item();
        }

        void set_model() {
            char x[20];
            sprintf(x, "st-shogun%d", int(get_holes()));
            Stone::set_model(x);
        }

        static Holes smallest_hole(Holes s);

        Holes transfer_smallest_hole(GridPos p);
        StoneResponse actor_hit(const StoneContact &sc);
    };
}

ShogunStone::Holes 
ShogunStone::get_holes() const 
{ 
    int h=int_attrib("holes"); 
    if (h>=1 && h<=7)
        return Holes(h);
    else {
        cerr << "Wrong `holes' attribute\n";
        return SMALL;
    }
}

ShogunStone::Holes 
ShogunStone::smallest_hole(Holes s) 
{
    if (s & SMALL)
        return SMALL;
    if (s & MEDIUM)
        return MEDIUM;
    if (!(s & LARGE))
        assert(0);
    return LARGE;
}

void 
ShogunStone::notify_item() 
{
    if (Item *it=GetItem(get_pos())) {
        switch (get_holes()) {
        case SMALL:
            SendMessage(it,"shogun1");
            break;
        case (MEDIUM | SMALL):
            SendMessage(it,"shogun2");
            break;
        case (LARGE | MEDIUM | SMALL):
            SendMessage(it,"shogun3");
            break;
        default:
            SendMessage(it, "noshogun");
            break;
        }
    }
}

ShogunStone::Holes
ShogunStone::transfer_smallest_hole(GridPos p)
{
    bool hole_transferred=false;
    Holes h = get_holes();
    Holes sh = smallest_hole(h);


    if (Stone *st=GetStone(p)) {
        if (ShogunStone *sst = dynamic_cast<ShogunStone*>(st)) {
            // Transfer the smallest hole to the other shogun stone.
            // Holes must be ``stacked'' properly
            if (sh < smallest_hole(sst->get_holes())) {
                sst->add_hole(sh);
                hole_transferred=true;
            }
        }
    }
    else {
        // empty field? Simply create a new stone on it
        ShogunStone *sst = new ShogunStone();
        sst->set_holes(sh);
        SetStone(p, sst);
        hole_transferred=true;
    }

    return (hole_transferred) ? Holes(h & ~sh) : h;
}

StoneResponse 
ShogunStone::actor_hit(const StoneContact &sc)
{
    Actor *a = sc.actor;
    bool move_impulse= (a->get_vel()*sc.normal < -4);
    Direction hit_dir=reverse(contact_face(sc));

    if (hit_dir == NODIR || !move_impulse)
        return STONE_REBOUND;

    GridPos newpos=move(get_pos(), hit_dir);
    Holes holes = get_holes();
    Holes newholes = transfer_smallest_hole(newpos);

    if (newholes!=holes) {
        sound::PlaySound("st-move");
        if (newholes == 0)
            return STONE_KILL;
        else {
            set_holes(newholes);
            notify_item();
            set_model();
        }
    }
    return STONE_REBOUND;
}


//----------------------------------------
// FartStone
//
// Messages:
//
// :trigger
//----------------------------------------
namespace
{
    class FartStone : public CloneStone<FartStone> {
    public:
        FartStone() : Stone("st-fart"), state(IDLE) {}
    private:
        enum State { IDLE, FARTING, BROKEN };
        State state;

        void fart() {
            if (state == IDLE) {
                Object *ox = world::GetObjectTemplate("st-oxyd");
                SendMessage(ox, "closeall");
                sound::PlaySound("fart");
                set_model("st-farting")
                    ->set_callback(px::make_cb(this, &FartStone::animcb));
                state = FARTING;
            }
        }
        void destroy() {

        }
        void animcb() { 
            if (state == FARTING) {
                state = IDLE;
                set_model(); 
            }
        }

        StoneResponse actor_hit(const StoneContact &sc) {
            if (wielded_item_is(sc.actor, "it-hammer")) {
                // ### What we really want is either a broken "fart"
                // ### stone or an animation
                return STONE_KILL;
            }
            fart();
            return STONE_REBOUND;
        }
        void message(const string &m, const Value &) {
            if (m=="trigger")
                fart();
        }
    };
}

//----------------------------------------
// Thief Stone
//
// Takes one item from inventory after actor hit.
//
// TODO: It should steal afer some period of time
//----------------------------------------
namespace
{
    class ThiefStone : public CloneStone<ThiefStone> {
    public:
        ThiefStone() : Stone("st-thief"), state(IDLE) { }
    private:
        enum State { IDLE, STEALING } state;

        StoneResponse actor_hit(const StoneContact &sc) {
            if (state==IDLE) {
                set_model("st-thief-anim")
                    ->set_callback(px::make_cb(this, &ThiefStone::animcb));
                if (player::Inventory *inv = get_inventory(sc.actor))
                {
                    int i = int(inv->size() * (rand()/(RAND_MAX+1.0)));
		    delete inv->yield_item(i);
                    
                    sound::PlaySound("thief");
                    state = STEALING;
                }
            }
            return STONE_REBOUND;
	}
        
        void animcb() {
            if (state == STEALING) {
                state = IDLE;
                set_model();
            }
        }

        const char *collision_sound() {
            return "st-thud";
        }
    };
}


//----------------------------------------
// ActorImpulseStone
//----------------------------------------
namespace
{
    class ActorImpulseStone : public CloneStone<ActorImpulseStone> {
    public:
        ActorImpulseStone() : Stone("st-actorimpulse"), state(IDLE) {}
    private:
	enum State { IDLE, PULSING, BROKEN };
	State state;

	px::V3
	actor_impulse(const StoneContact &sc)
	{
	    V3 normal = normalize(sc.normal);
            const V3 &v = sc.actor->get_vel();
	    return -2*(v*normal)*normal;
	}

	
	StoneResponse actor_hit(const StoneContact &sc) {
            if (state == IDLE) {
//	        sound::PlaySound("stoneimpulse");
                set_model("st-actorimpulse-anim")
                    ->set_callback(px::make_cb(this, &ActorImpulseStone::animcb));
                state = PULSING;
            }
	    return STONE_REBOUND;
	}

        void animcb() {
            if (state == PULSING) {
                state = IDLE;
                set_model();
            }
        }
    };
}


//----------------------------------------
// StoneImpulseStone
//
// Messages:
//
// :trigger
//----------------------------------------
namespace
{
    class StoneImpulseStone : public CloneStone<StoneImpulseStone> {
    public:
        StoneImpulseStone() : Stone("st-stoneimpulse"), state(IDLE) {}
    private:
        enum State { IDLE, PULSING, BROKEN };
        State state;

        void stoneimpulse() {
            if (state == IDLE) {
//                sound::PlaySound("stoneimpulse");
                set_model("st-stoneimpulse-anim")
                    ->set_callback(px::make_cb(this, &StoneImpulseStone::animcb));
                state = PULSING;
                MaybeMoveStone( move( get_pos(), NORTH), NORTH);
                MaybeMoveStone( move( get_pos(), EAST), EAST);
                MaybeMoveStone( move( get_pos(), SOUTH), SOUTH);
                MaybeMoveStone( move( get_pos(), WEST), WEST);
            }
        }
        void animcb() {
            if (state == PULSING) {
                state = IDLE;
                set_model();
            }
        }

        StoneResponse actor_hit(const StoneContact &sc) {
            stoneimpulse();
            return STONE_REBOUND;
        }
        void message(const string &m, const Value &) {
            if (m=="trigger")
                stoneimpulse();
        }
    };
}

//----------------------------------------
// OxydStone
//----------------------------------------

/** \page st-oxyd Oxyd Stone

Oxyd stones are characterized by two attributes: Their flavor and
their color.  The flavor only affects the visual representation of
the stone; it can be either 'a' (opening like a flower) or 'b'
(displaying a fade-in animation).  The color attribute determines
the shape on the oxyd stone.

\b Note: You should usually not to create Oxyd stones manually with \c set_stone(). 
Use the predefined \c oxyd() function instead.

\subsection oxyda Attributes

- \b flavor      "a", "b", "c", or "d"
- \b color       number between 0 and 7

\subsection oxydm Messages

- \b closeall    close all oxyd stones
- \b shuffle     interchange the colors of the oxyd stones in the current landscape
- \b trigger     open the stone

<table><tr>
<td>\image html st-oxyda.png "flavor A"
<td>\image html st-oxydb.png "flavor B"
<td>\image html st-oxydc.png "flavor C"
<td>\image html st-oxydd.png "flavor D"
</table>
*/

namespace
{
    class OxydStone : public laser::PhotoStone {
    public:
        OxydStone();
        ~OxydStone();

        static void shuffle_colors();
    private:
        enum State { CLOSED, OPEN, OPENING, CLOSING, BLINKING };
        State state;

        display::Model *model_bak;
        typedef vector<OxydStone*> InstanceList;
        static InstanceList instances;

        static bool blinking_or_opening(OxydStone *a) { 
            return (a->state==BLINKING || a->state==OPENING); 
        }
        static bool not_open(OxydStone *a) {
            return a->state!=OPEN;
        }

        // Stone interface
        StoneResponse actor_hit(const StoneContact &sc);
        void on_creation();
        void on_removal();
        const char *collision_sound() { return "st-metal"; }
        void message(const string &m, const Value &);


        // Object interface
        Object *clone();
        void dispose();

        // Animation callback
        void animcb();

        void maybe_open_stone();
        void change_state(State newstate);

        void notify_laseron() { maybe_open_stone(); }
        void notify_laseroff() {}
    };
}

vector<OxydStone*> OxydStone::instances;

OxydStone::OxydStone() 
    : Stone("st-oxyd"), state(CLOSED), model_bak(0)
{
    set_attrib("flavor", "b");
    set_attrib("color", "0");
}

OxydStone::~OxydStone()
{
    delete model_bak;
}

void 
OxydStone::message(const string &m, const Value &)
{
    if (m=="closeall") {
        for (unsigned i=0; i<instances.size(); ++i) {
            instances[i]->change_state(CLOSING);
        }
    } 
    else if (m=="shuffle") {
        shuffle_colors();
    }
    else if (m=="trigger") {
	maybe_open_stone();
    }
}

void 
OxydStone::shuffle_colors()
{
    for (unsigned i=0; i<instances.size(); ++i) {
        int a=int(instances.size() * (rand()/(RAND_MAX+1.0)));
        string icolor=instances[i]->string_attrib("color");
        string acolor=instances[a]->string_attrib("color");
        instances[i]->set_attrib("color", acolor.c_str());
        instances[a]->set_attrib("color", icolor.c_str());
    }
}

void
OxydStone::change_state(State newstate)
{
    string modelname = "st-oxyd";
    modelname += string_attrib("flavor");
    modelname += string_attrib("color");

    switch (newstate) {
    case CLOSED: 
        Stone::set_model(string("st-oxyd")+string_attrib("flavor"));
        break;

    case BLINKING:
        Stone::set_model(modelname + "-blink");
        break;
	
    case OPEN:
	if (state == CLOSED) {
            sound::PlaySound("st-oxydopen");
            sound::PlaySound("st-oxydopened");
            Stone::set_model(modelname+"-opening")
                ->set_callback(px::make_cb(this, &OxydStone::animcb));
        } else {
            Stone::set_model(modelname + "-open");

            // If this was the last closed oxyd stone, finish the
            // level
            if (find_if(instances.begin(),instances.end(),not_open)
                ==instances.end())
            {
                enigma::FinishLevel();
            }
        }
        break;
	
    case OPENING:
        sound::PlaySound("st-oxydopen");
	if (state == CLOSED) {
            Stone::set_model(modelname + "-opening")
                ->set_callback(px::make_cb(this, &OxydStone::animcb));
	}
	else if (state == CLOSING)
            get_model()->reverse();
	break;
	
    case CLOSING:
        if (state == OPENING)
            get_model()->reverse();
	else if (state == BLINKING || state == OPEN) {
            Stone::set_model(modelname + "-closing")
                ->set_callback(px::make_cb(this, &OxydStone::animcb));
        }
        else if (state == CLOSED || state==CLOSING)
            return; // do nothing
        break;
    }
    state = newstate;
}

void 
OxydStone::animcb()
{
    if (state == CLOSING)
        change_state(CLOSED);
    else if (state == OPENING)
        change_state(BLINKING);
    else if (state == OPEN)
        change_state(OPEN); // set the right model
}

void
OxydStone::maybe_open_stone()
{
    if (state == CLOSED || state == CLOSING) {
        int mycolor = int_attrib("color");

        // Is another oxyd stone currently blinking?
        InstanceList::iterator i;
        i=find_if(instances.begin(), instances.end(), blinking_or_opening);

        if (i != instances.end()) {
            // If colors match, open both stones. Close one of them
            // otherwise
            if (mycolor == (*i)->int_attrib("color")) {
                change_state(OPEN);
                (*i)->change_state(OPEN);
            } else {
                (*i)->change_state(CLOSING);
                change_state(OPENING);
            }
        } 
        else {
            // no blinking stone? -> make this one blink
            change_state(OPENING);
        }
    }
}

StoneResponse 
OxydStone::actor_hit(const StoneContact &sc)
{
    maybe_open_stone();
    return STONE_REBOUND;
}

void 
OxydStone::on_creation()
{
    if (model_bak) {
        display::SetModel(GridLoc(GRID_STONES, get_pos()), model_bak);
        model_bak=0;
    } 
    else
        set_model(string("st-oxyd") + string_attrib("flavor"));
    
    photo_activate();
}

void 
OxydStone::on_removal()
{
    photo_deactivate();
    model_bak = display::YieldModel(GridLoc(GRID_STONES, get_pos()));
}

Object*
OxydStone::clone() {
    OxydStone *o = new OxydStone;
    instances.push_back(o);
    return o;
}

void
OxydStone::dispose() {
    instances.erase(find(instances.begin(),
                         instances.end(), this));
    delete this;
}



//----------------------------------------
// FakeOxydStone
//----------------------------------------

/** \page st-fakeoxyd Fake Oxyd Stone

These stones look like real Oxyd stones, but they only blink a little
when touched and do not open or have other special abilities.

\image html st-fakeoxyd-blink_0001.png
*/
namespace
{
    class FakeOxydStone : public CloneStone<FakeOxydStone> {
    public:
        FakeOxydStone() : Stone("st-fakeoxyd"), state(IDLE) {}
    private:
        enum State { IDLE, BLINKING } state;
        StoneResponse actor_hit(const StoneContact &sc) {
            if (state == IDLE) {
                set_model("st-fakeoxyd-blink")
                    ->set_callback(px::make_cb(this, &FakeOxydStone::animcb));
                state = BLINKING;
            }
            return STONE_REBOUND;
        }
        const char *collision_sound() {
            return "st-fakeoxyd";
        }
        void animcb() {
            set_model("st-fakeoxyd");
            state = IDLE;
        }
    };
}

//----------------------------------------
// Grate1
//----------------------------------------
namespace
{
    class Grate1 : public CloneStone<Grate1> {
    public:
        Grate1() : Stone("st-grate1") {}
        StoneResponse actor_hit(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// Grate2
//----------------------------------------
namespace
{
    class Grate2 : public CloneStone<Grate2> {
    public:
        Grate2() : Stone("st-grate2") {}
        StoneResponse actor_hit(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// Marble_hole
//----------------------------------------
namespace
{
    class Marble_hole : public CloneStone<Marble_hole> {
    public:
        Marble_hole() : Stone("st-marble_hole") {}
        StoneResponse actor_hit(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// Greenbrown_hole
//----------------------------------------
namespace
{
    class Greenbrown_hole : public CloneStone<Greenbrown_hole> {
    public:
        Greenbrown_hole() : Stone("st-greenbrown_hole") {}
        StoneResponse actor_hit(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// Rock3_hole
//----------------------------------------
namespace
{
    class Rock3_hole : public CloneStone<Rock3_hole> {
    public:
        Rock3_hole() : Stone("st-rock3_hole") {}
        StoneResponse actor_hit(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// BombStone
//
// These stones add a bomb to the player's inventory when touched.
//----------------------------------------
namespace
{
    class BombStone : public CloneStone<BombStone> {
    public:
        BombStone() : Stone("st-bombs") {}
    private:
        StoneResponse actor_hit(const StoneContact &sc);
    };
}

StoneResponse 
BombStone::actor_hit(const StoneContact &sc)
{
    if (player::Inventory *inv = get_inventory(sc.actor))
    {
        if (!inv->is_full())
        {
            Item *it = dynamic_cast<Item*>(MakeObject("it-blackbomb"));
            inv->add_item(it);
        }
    }
    return STONE_REBOUND;
}

//----------------------------------------
// GlassStone
//
// Laser light may pass this stone, but actors may not.
//----------------------------------------
namespace
{
    class GlassStone : public CloneStone<GlassStone> {
    public:
        GlassStone() : Stone("st-glass") {}
    private:
        bool on_laserhit(Direction dir) { return true; }
    };
}

//----------------------------------------
// DeathStone
//----------------------------------------

/** \page st-death Death's Head Stone

Simply kills all actors that touch it (except for actors that are
immune to these stones).

\image html st-death.png
*/
namespace
{
    class DeathStone : public CloneStone<DeathStone> {
    public:
        DeathStone() : Stone("st-death"), active(false) {}
    private:
        bool active; 
        StoneResponse actor_hit(const StoneContact &sc) 
        {
	    if (!active) {
		SendMessage(sc.actor, "shatter");
		active=true;
		set_model("st-death-munch")
		    ->set_callback(px::make_cb(this, &DeathStone::animcb));
	    }
            return STONE_REBOUND;
        }
	void animcb() { set_model("st-death"); active=false; }
    };
}


//----------------------------------------
// Coin slot Switch
//----------------------------------------

/** \page st-coinslot Coin Slot Stone 

Coin Slots are switches that must be activated by inserting coins.
The more coins you put in, the longer the switch will stay activated

\subsection coinslota Attributes

- \b cost            how many on-time bought for coin1
- \b interval        on-time remaining
- \b target, \b action   as usual

\subsection coinslote Example

\verbatim
set_stone("st-coinslot", 20,13,{action="openclose", target="door", cost=1})
\endverbatim
*/

namespace
{
    class CoinSlot : public CloneStone<CoinSlot>,
                     public OnOffStone,
                     public tools::TimeHandler {
    public:
        CoinSlot();
    private:
        // Variables.
        enum State { ACTIVE, INACTIVE } state;
        double remaining_time;

        // Private methods.
        void set_model() {
            Stone::set_model(state==ACTIVE ? "st-coinslot" : "st-coinslot"); 
            // st-coinslot-filled
        }

        StoneResponse actor_hit(const StoneContact &sc);
        const char *collision_sound() { return "st-metal"; }

        void animcb() { set_model(); }
        void tick(double dtime);
        void change_state(State newstate);
    };
}

CoinSlot::CoinSlot() 
    : Stone("st-coinslot"), state(INACTIVE), remaining_time(0)
{
    set_attrib("cost", 1.0);
}

void
CoinSlot::change_state(State newstate)
{
    if (state == newstate) return;

    switch (newstate) {
    case ACTIVE:
        PerformAction(this, true);
        g_timer.activate(this);
        break;
    case INACTIVE:
        PerformAction(this, false);
        g_timer.deactivate(this);
        break;
    }
    state = newstate;
}

void
CoinSlot::tick(double dtime)
{
    assert(remaining_time > 0);
    
    remaining_time -= dtime;
    if (remaining_time <= 0)
    {
        change_state(INACTIVE);
    }
}

StoneResponse 
CoinSlot::actor_hit(const StoneContact &sc) 
{
    if (wielded_item_is(sc.actor, "it-coin")) 
    {
        if (player::Inventory *inv = get_inventory(sc.actor))
        {
            Item *it = inv->yield_first();
            remaining_time += it->double_attrib("value")*double_attrib("cost");
            Stone::set_model("st-coin2slot")
                 ->set_callback(px::make_cb(this, &CoinSlot::animcb));
            change_state(ACTIVE);
            delete it;
        }
    }
    return STONE_REBOUND;
}

//----------------------------------------------------------------------
// OBJECT REPOSITORY
//----------------------------------------------------------------------
namespace
{
    class ObjectRepos : public px::Nocopy {
    public:
        ObjectRepos();
        ~ObjectRepos();
        void add_templ(Object *o);
        bool has_templ(const string &name);
        Object *make(const string &name);
        Object *get_template(const string &name);
    private:
        typedef map<string, Object*> ObjectMap;
        ObjectMap objmap;           // repository of object templates
        int stonecount, floorcount, itemcount;
    };
}

ObjectRepos::ObjectRepos()
{
    stonecount=floorcount=itemcount=0;

    // Floors
    add_templ(new Abyss);
    add_templ(new Water);
    add_templ(new Bridge);
    add_templ(new Gradient);

    // Stones
    add_templ(new BolderStone);
    add_templ(new BombStone);
    add_templ(new BrickMagic);
    add_templ(new Brownie);
    add_templ(new ChameleonStone);
    add_templ(new CoinSlot);
    add_templ(new DeathStone);
    add_templ(new Door);
    add_templ(new Door_a);
    add_templ(new Door_b);
    add_templ(new Door_c);
    add_templ(new FakeOxydStone);
    add_templ(new FartStone);
    add_templ(new FloppyStone);
    add_templ(new GlassStone);
    add_templ(new Grate1);
    add_templ(new Grate2);
//    add_templ(new KeyStone);
    add_templ(new KeyStone_a);
    add_templ(new KeyStone_b);
    add_templ(new KeyStone_c);
    add_templ(new OneWayStone);
    add_templ(new OxydStone);
    add_templ(new PuzzleStone);
    add_templ(new ShogunStone);
    add_templ(new ActorImpulseStone);
    add_templ(new StoneImpulseStone);
    add_templ(new SwapStone);
    add_templ(new SwitchStone);
    add_templ(new ThiefStone);
    add_templ(new TimerStone);
    add_templ(new WoodenStone);

    add_templ(new Stone_break);
    add_templ(new Greenbrown_hole);
    add_templ(new Greenbrown_move);
    add_templ(new Marble_hole);
    add_templ(new Marble_move);
    add_templ(new Rock3_hole);
    add_templ(new Rock3_move);
}

ObjectRepos::~ObjectRepos()
{
    px::delete_map(objmap.begin(), objmap.end());
}

// Add an new Object template to `objmap'. 
void 
ObjectRepos::add_templ(Object *o)
{
    string kind = o->get_kind();
    if (has_templ(kind))
        cerr << "add_templ: redefinition of object `" <<kind<< "'.\n";
    else
        objmap[kind] = o;
}

bool 
ObjectRepos::has_templ(const string &name)
{
    return objmap.find(name) != objmap.end();
}

Object *
ObjectRepos::make(const string &name)
{
    return objmap[name]->clone();
}

Object *
ObjectRepos::get_template(const string &name)
{
    return objmap[name];
}

//----------------------------------------------------------------------
// FUNCTIONS
//----------------------------------------------------------------------
namespace
{
    ObjectRepos repos;
}

void 
world::Register(Object *obj)
{
    repos.add_templ(obj);
}

Object *
world::MakeObject(const string &kind)
{
    if (!repos.has_templ(kind)) {
        cerr << "MakeObject: unkown object name `" <<kind<< "'.\n";
        return 0;
    } else
        return repos.make(kind);
}

Object *
world::GetObjectTemplate(const std::string &kind)
{
    if (!repos.has_templ(kind)) {
        cerr << "GetObjectTemplate: unkown object name `" <<kind<< "'.\n";
        return 0;
    } else
        return repos.get_template(kind);
}


Floor*
world::MakeFloor(const std::string &kind)
{
    return dynamic_cast<Floor*>(MakeObject(kind));
}

Item *
world::MakeItem(const string &kind)
{
    return dynamic_cast<Item*>(MakeObject(kind));
}

void
world::DisposeObject(Object *o)
{ 
    if (o != 0) {
        UnnameObject(o);
        o->dispose();
    }
}

void 
world::SendMessage(Object *o, const std::string &msg)
{
    if (o)
        o->message(msg, Value());
}

void 
world::DefineSimpleStone(const std::string &kind, 
                         const std::string &sound)
{
    Register(new SimpleStone(kind,sound));
}

void 
world::DefineSimpleFloor(const std::string &kind,
                         double friction, double mousefactor)
{
    Register(new Floor(kind, friction, mousefactor));
}
