//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#ifndef OBJECTS_HH
#define OBJECTS_HH

//#include "fwd.hh"
#include "world.hh"
#include "display.hh"
#include "tools.hh"

#include "px/pxfwd.hh"
#include "px/math.hh"
#include <string>

//
// Object is the base class for all ``objects'' in the world.  The
// most important facilities this class provides are:
//
// (1) A way to clone() and dispose() objects.  This is mainly used in
//     function MakeObject() to create new objects of a given type.
//
// (2) A way to pass messages between unrelated objects via message().
//     This allows us to send messages to objects from Lua and to
//     decouple objects types as much as possible.
//
// (3) A way to get and set attributes.  These attributes are quite
//     similar to instance variables, but they can be easily modified
//     from Lua.  This makes it possible to modify certain object
//     parameters (such as the text on a piece of paper or the color
//     of an oxyd stone) in level descriptions.
//

namespace world
{
    using tools::Value;
    using std::string;

    class UndefinedAttrib {
    public:
        const char *what() const { return "undefined attribute"; }
    };
        
    
    class Object {
    public:
        Object() : kind_(), refcnt(0) {}
        Object(const string& kind);
        virtual ~Object() {}

        // Object interface
        virtual void message(const string& msg, const Value &val) {}
        virtual void set_attrib(const string& key, const Value &val);
        virtual const Value* get_attrib(const string& key) const;

        const char *string_attrib(const string &name) const;
        int int_attrib(const string &name) const;
        double double_attrib(const string &name) const;

        // Increase and decrease reference count
        void incref() { refcnt++; }
        void decref() { if (--refcnt == 0) delete this; }

        virtual Object *clone()=0;
        virtual void dispose()=0;

        const string &get_kind() const { return kind_; }
    private:
        string kind_;      // the kind of object this is
        tools::ValueMap attribs;
        unsigned refcnt;
    };
}


// GridObject is the base class for everything that can only be placed
// on "The Grid".
namespace world
{
    class GridObject : public Object {
    public:
        GridObject() {}
        GridObject(const string kind) : Object(kind) {}

        void creation(GridPos p);
        void removal() { on_removal(); }

        virtual bool on_laserhit(Direction d) { return false; }

        // GridObject interface
        virtual void actor_enter(Actor *a) {}
        virtual void actor_leave(Actor *a) {}

        GridPos get_pos() const { return pos; }

    private:
        virtual void on_creation() = 0;
        virtual void on_removal() = 0;
        GridPos pos;
    };
}

//
// Floor
//
namespace world
{
    class Floor : public GridObject {
    public:
        Floor(const string &kind, double friction, double mfactor);

        // Object interface
        Object *clone();
        void dispose();
        void message(const string& msg, const Value &val);


        // Floor interface
        virtual px::V3 get_force(Actor *a)
        { return px::V3(); }

        virtual void on_drop(Item* it) {}
        virtual void on_pickup(Item* it) {}

        virtual void stone_change(Stone *st) {}

        void actor_hit(Actor *a) { on_actorhit(a); }

        double friction() const { return _friction; }
        double mousefactor() const { return _mousefactor; }
    protected:
        Floor() {}
        display::Model * set_model(const string& name);
        display::Model * get_model();
    private:
        // GridObject interface
        virtual void on_creation();
        virtual void on_removal();

        virtual void on_actorhit(Actor *a) {}

        double _friction;
        double _mousefactor;
    };
}

//
// Item
//
namespace world
{
    // What may happen to an item _after_ it was activated?
    enum ItemAction {
        ITEM_DROP, // drop it to the floor
        ITEM_KILL, // remove it from the inventory and dispose it
        ITEM_KEEP  // keep it in the inventory; do nothing further
    };

    class Item : public GridObject {
    public:
        Item(const string &kind);

        // Item interface
        virtual px::V3 get_force(Actor *a)
        { return px::V3(); }

        virtual void on_drop(Actor *a) {}
        virtual void on_pickup(Actor *a) {}

        virtual void stone_change(Stone *st) {}

        // Return true if the item should be picked up.
        virtual bool actor_hit(Actor *a);

        // The model used for displaying this item in an inventory 
        virtual string get_inventory_model();

        // Item is activated by the player
        virtual ItemAction activate(Actor* a, GridPos p)
        { return ITEM_DROP; }

    protected:
        Item() {}
        display::Model * set_model(const string &name);
    private:
        virtual void on_creation();
        virtual void on_removal();
    };
}

//
// Stone
//
namespace world
{
    class Stone : public GridObject {
    public:
        Stone(const string & kind);

        // Stone interface
        virtual px::V3          actor_impulse (const StoneContact &sc);
        virtual StoneResponse   actor_hit (const StoneContact &sc);
        virtual void            actor_inside (Actor *a) {}
        virtual void            actor_contact (Actor *a) {}
        virtual bool            on_move() { return false;}
        virtual const char *    collision_sound();
    protected:
        Stone() {}
        display::Model * set_model(const string &mname);
        display::Model * get_model();

        virtual void set_model();

        virtual void on_creation();
        virtual void on_removal();
    };
}

namespace world
{
    void SendMessage(Object *o, const string &msg);

    // This function is used by all triggers, switches etc. that perform
    // some particular action when activated (like opening doors or
    // switching lasers on and off).
    void PerformAction(Object *o, bool onoff);



    Object *MakeObject(const string &kind);
    Floor *MakeFloor(const string &kind);
    Item *MakeItem(const string &kind);
    void DisposeObject(Object *o);

    void DefineSimpleStone(const string &kind, const string &sound);
    void DefineSimpleFloor(const string &kind,
                           double friction, double mousefactor);

    // Register a new object under the name obj->kind().  Objects
    // registered in this way can be created with MakeObject(), where
    // `obj' is used as a template.  
    void Register(Object *obj);

    Object *GetObjectTemplate(const string &kind);
}

#endif
