//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================

// General video access not tied to any particular game state.  See
// display.hh for the in-game video functionality.

#include "enigma.hh"
#include "video.hh"
#include "lua.hh"
#include "px/px.hh"
#include "SDL.h"
#include <cassert>
#include <cstdio>

using namespace std;
using namespace px;

//----------------------------------------
// Video engine
//----------------------------------------
namespace
{
    class Video_SDL {
        SDL_Surface*    sdlScreen;
        bool            fullScreen;
        string          caption;
        px::Screen*     screen;
        bool            initialized;
    public:
        Video_SDL();
        ~Video_SDL();

        void init();
        void toggle_inputgrab();
        void toggle_fullscreen();
        void set_fullscreen(bool on_off);
        bool is_fullscreen() const { return fullScreen; }
        void set_caption(const char *str);
        px::Screen *get_screen() { return screen; }
    };
}

Video_SDL::Video_SDL()
    : sdlScreen(0), fullScreen(false), screen(0), initialized(false)
{}

Video_SDL::~Video_SDL()
{
    SDL_WM_GrabInput(SDL_GRAB_OFF);
    if (sdlScreen != 0 && fullScreen)
        SDL_WM_ToggleFullScreen(sdlScreen);
    delete screen;
}

void Video_SDL::set_caption(const char *str)
{
    if (initialized)
        SDL_WM_SetCaption(str, 0);
}

void Video_SDL::init()
{
    SDL_WM_SetCaption(caption.c_str(), 0);

    int w = 640;
    int h = 480;
    int bpp = 16;

    sdlScreen = SDL_SetVideoMode(w, h, bpp, SDL_FULLSCREEN);
    if (sdlScreen == 0)
    {
	fprintf(stderr, "Couldn't open screen: %s\n", SDL_GetError());
        exit(1);
    }
    if (sdlScreen->flags & SDL_FULLSCREEN)
        fullScreen = true;
    screen = px::CreateScreen(sdlScreen);
    assert(screen != 0);

    SDL_ShowCursor(0);
    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY / 2,
                        SDL_DEFAULT_REPEAT_INTERVAL / 2);

    initialized = true;
}

void Video_SDL::toggle_inputgrab()
{
    if (SDL_WM_GrabInput(SDL_GRAB_QUERY) == SDL_GRAB_ON)
        SDL_WM_GrabInput(SDL_GRAB_OFF);
    else
        SDL_WM_GrabInput(SDL_GRAB_ON);
}

void Video_SDL::set_fullscreen(bool on_off)
{
    if (on_off != fullScreen)
    {
        toggle_fullscreen();
        fullScreen = on_off;
    }
}

void Video_SDL::toggle_fullscreen()
{
    SDL_WM_ToggleFullScreen(sdlScreen);
    fullScreen = !fullScreen;
}

//----------------------------------------
// MouseCursor
//----------------------------------------
namespace
{
    class MouseCursor {
    public:
        MouseCursor()
	    : background(0), cursor(0)
	{
            oldx = oldy = 0;
	    hotx = hoty = 0;
	    visible = 0;
	    changedp = true;
        }
        ~MouseCursor() {
            delete background;
        }

        void set_image(px::Surface *s, int hx, int hy)
        {
            cursor=s;
            hotx=hx;
            hoty=hy;
        }
	
        void redraw(Screen *scr) {
	    if (visible>0 && cursor) {
		GC gc(scr);
		int scrx=x-hotx;
		int scry=y-hoty;

                if (background) {
                    scr->blit(oldx-hotx, oldy-hoty, background);
                    scr->update_rect(get_oldrect());
                }
                grab_bg(scr);

		blit(gc, scrx, scry, cursor);
		scr->update_rect(get_rect());

                oldx=x;
                oldy=y;
	    }
	    changedp=false;	    
        }
        void move(int newx, int newy) {
            x=newx;
            y=newy;
	    changedp=true;
        }
        void show(Screen *scr) {
	    if (++visible == 1)
		changedp=true;
        }
        void hide(Screen *scr) {
	    if (--visible == 0) {
		changedp=true;
                if (background)
                    scr->blit(oldx-hotx, oldy-hoty, background);
                delete background;
                background=0;
            }
        }
        Rect get_rect() const {
            int scrx=x-hotx;
            int scry=y-hoty;
            return Rect(scrx,scry,cursor->width(), cursor->height());
        }
        Rect get_oldrect() const {
            int scrx=oldx-hotx;
            int scry=oldy-hoty;
            return Rect(scrx,scry,cursor->width(), cursor->height());
        }

	bool has_changed() { return changedp; }
        int get_x() const { return x; }
        int get_y() const { return y; }
        
    private:
        void grab_bg(Screen *scr) {
            if (background == 0)
                background = Grab(scr, get_rect());
            else 
                background->blit(0,0, scr, get_rect());
        }

        Surface *background;
        Surface *cursor;
        int x, y;
        int oldx, oldy;
	int hotx, hoty;
	int visible;
	bool changedp;
    };
}




//----------------------------------------------------------------------
// FUNCTIONS.
//----------------------------------------------------------------------
namespace
{
    Video_SDL *video_engine;
    MouseCursor cursor;
}

static int
event_filter(const SDL_Event *e)
{
    if (e->type == SDL_MOUSEMOTION)
    {
        cursor.move(e->motion.x, e->motion.y);
        cursor.redraw(video::GetScreen());
    }
    else if (e->type == SDL_KEYDOWN)
    {
        if ((e->key.keysym.sym==SDLK_RETURN) &&
            (e->key.keysym.mod & KMOD_ALT)) 
        {
            video::ToggleFullscreen();
            return 0;
        }
    }
    return 1;
}

void 
video::SetMouseCursor(px::Surface *s, int hotx, int hoty)
{
    cursor.set_image(s, hotx, hoty);
    cursor.redraw(video::GetScreen());
}

void 
video::HideMouse()
{
    cursor.hide(GetScreen());
    cursor.redraw(video::GetScreen());
}

void 
video::ShowMouse()
{
    cursor.show(GetScreen());
    cursor.redraw(video::GetScreen());
}

int 
video::Mousex()
{
    return cursor.get_x();
}

int 
video::Mousey()
{
    return cursor.get_y();
}



void
video::Init()
{
    video_engine = new Video_SDL();
    video_engine->init();

    int x, y;
    SDL_GetMouseState(&x, &y);
    cursor.move(x,y);


    SDL_SetEventFilter(event_filter);

    lua::set_num("FullScreen", video_engine->is_fullscreen());

}

void
video::Shutdown()
{
    delete video_engine;
}


px::Screen *
video::GetScreen()
{
    return video_engine->get_screen();
}


void 
video::ToggleFullscreen() 
{
    video_engine->toggle_fullscreen();    
    lua::set_num("FullScreen", video_engine->is_fullscreen());
}

void
video::ToggleInputgrab() {
    ; //video_engine->toggle_inputgrab();    
}

void
video::SetCaption(const char *str)
{
    video_engine->set_caption(str);
}

void 
video::Fade(FadeMode mode)
{
    const double fadesec = 0.8;
    double v = 255/fadesec;
        
    px::Screen *screen = GetScreen();
    px::Surface *buffer = Duplicate(screen);
    double dt;

    double a = mode==FADEIN ? 0 : 255;
    while (true) {
        Uint32 otime = SDL_GetTicks();

        screen->box(0,0,640,480, 0);
        buffer->set_alpha(int(a));
        screen->blit(0,0,buffer);
        screen->update_all();
        screen->flush_updates();

        dt = (SDL_GetTicks()-otime)/1000.0;
        if (mode==FADEIN && (a+=v*dt) > 255)
            break;
        else if (mode==FADEOUT && (a-=v*dt) < 0)
            break;
    }

    if (mode==FADEIN) {
        buffer->set_alpha(255);
        screen->blit(0,0,buffer);
    } else
        screen->box(0,0,640,480, 0);
    screen->update_all();
    screen->flush_updates();
    delete buffer;
}


void 
video::Screenshot(const char *fname)
{
    SDL_SaveBMP(GetScreen()->get_surface(), fname);
}


void 
video::DisplayTextCentered(const char *fontname, 
                           const char *text)
{
    Font &f = enigma::GetFont(fontname);
    Screen *s = GetScreen();
    s->box(0,0,s->width(),s->height(), 0);
    f.render(s,
             (s->width() - f.get_width(text))/2,
             (s->height() - f.get_height())/2,
             text);
    s->update_all();
    s->flush_updates();
}
